import datetime
import time
from unittest import mock

import pytest
from django.db import models
from django.test import override_settings
from django.utils import timezone

from example_app.models import GoalRelatedModel

from .blocking_worker import listen_goal_waiting_for_worker
from .factories import GoalFactory, GoalProgressFactory
from .models import (
    AllDone, Goal, GoalState, PreconditionFailureBehavior, PreconditionsMode,
    RetryMeLater, RetryMeLaterException, handle_waiting_for_preconditions,
    handle_waiting_for_worker, schedule, thread_local, worker, worker_turn,
)


@pytest.mark.django_db
def test_worker_turn_noop():
    now = timezone.now()
    transitions_done = worker_turn(now)
    assert transitions_done == (0, 0)


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [
    {'state': GoalState.WAITING_FOR_DATE},
    {'state': GoalState.WAITING_FOR_PRECONDITIONS},
    {'state': GoalState.WAITING_FOR_WORKER},
    {'state': GoalState.ACHIEVED},
    {'state': GoalState.GIVEN_UP},
], indirect=True)
def test_handle_waiting_for_worker_return_value(goal):
    progress = handle_waiting_for_worker()
    did_a_thing = progress is not None
    assert did_a_thing is (goal.state == GoalState.WAITING_FOR_WORKER)


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'handler': 'os.path.join',
    'instructions': {
        'args': [1, 2],
        'kwargs': {'a': 'b'},
    },
}], indirect=True)
def test_handle_waiting_for_worker_success(goal):
    dependent_goal = GoalFactory(
        waiting_for_count=42,
        precondition_goals=[goal],
    )

    with mock.patch('os.path.join') as func:
        func.return_value = {'aaa': 'im happy'}  # will be ignored
        handle_waiting_for_worker()

    assert func.call_count == 1
    assert func.call_args == mock.call(goal, 1, 2, a='b')

    goal.refresh_from_db()
    assert goal.state == GoalState.ACHIEVED

    progress = goal.progress.get()
    assert progress.success
    assert progress.time_taken > datetime.timedelta(0)

    # dependent goal is updated
    dependent_goal.refresh_from_db()
    assert dependent_goal.waiting_for_count == 41


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'precondition_date': timezone.now() - timezone.timedelta(days=1),
}], indirect=True)
def test_handle_waiting_for_worker_failure(goal):
    now = timezone.now()
    with mock.patch('django_goals.models.follow_instructions') as follow_instructions:
        follow_instructions.side_effect = Exception
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    assert goal.precondition_date > now

    progress = goal.progress.get()
    assert not progress.success


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'waiting_for_count': 0,
}], indirect=True)
def test_handle_waiting_for_worker_retry(goal):
    other_goals = GoalFactory.create_batch(2, state=GoalState.WAITING_FOR_WORKER)
    failed_goals = GoalFactory.create_batch(1, state=GoalState.NOT_GOING_TO_HAPPEN_SOON)
    precondition_date = timezone.now() + timezone.timedelta(days=1)
    with mock.patch('django_goals.models.follow_instructions') as follow_instructions:
        follow_instructions.return_value = RetryMeLater(
            precondition_date=precondition_date,
            precondition_goals=other_goals + failed_goals,
        )
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    assert goal.precondition_date == precondition_date
    assert set(goal.precondition_goals.all()) == set(other_goals + failed_goals)
    assert goal.waiting_for_count == 3
    assert goal.waiting_for_failed_count == 1

    progress = goal.progress.get()
    assert progress.success


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'waiting_for_count': 0,
}], indirect=True)
@pytest.mark.parametrize('already_present', [True, False])
def test_handle_waiting_for_worker_retry_precond_already_present(goal, already_present):
    precondition_goal = GoalFactory(state=GoalState.ACHIEVED)
    if already_present:
        goal.precondition_goals.add(precondition_goal)
    with mock.patch('django_goals.models.follow_instructions') as follow_instructions:
        follow_instructions.return_value = RetryMeLater(
            precondition_goals=[precondition_goal],
        )
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    assert goal.precondition_goals.get() == precondition_goal
    assert goal.waiting_for_count == 0


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'handler': 'os.path.join',
}], indirect=True)
def test_handle_waiting_for_worker_retry_by_exception(goal):
    precondition_date = timezone.now() + timezone.timedelta(days=1)
    with mock.patch('os.path.join') as handler:
        handler.side_effect = RetryMeLaterException(
            precondition_date=precondition_date,
            message='asdf',
        )
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    assert goal.precondition_date == precondition_date

    progress = goal.progress.get()
    assert progress.success
    assert progress.message == 'asdf'


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'handler': 'os.path.join',
    'preconditions_mode': PreconditionsMode.ANY,
    'waiting_for_count': 0,
    'waiting_for_not_achieved_count': 1,
}], indirect=True)
@pytest.mark.parametrize(
    ('precondition_goals', 'expected_waiting_for_count'),
    [
        ([], 1),
        (None, 0),  # means "retry immediately"
    ],
)
def test_handle_waiting_for_worker_any_mode_retry(goal, precondition_goals, expected_waiting_for_count):
    """
    In ANY preconditions mode, we should increment waiting_for_count
    even if RetryMeLater does not contain any new goals.
    """
    achieved_precond = GoalFactory(state=GoalState.ACHIEVED)
    waiting_precond = GoalFactory(state=GoalState.WAITING_FOR_DATE)
    goal.precondition_goals.add(achieved_precond, waiting_precond)

    with mock.patch('os.path.join') as handler:
        handler.return_value = RetryMeLater(precondition_goals=precondition_goals)
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    assert goal.waiting_for_count == expected_waiting_for_count
    assert goal.waiting_for_not_achieved_count == 1


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('precond_state', 'expected_waiting_for_count'),
    [
        (GoalState.ACHIEVED, 0),
        (GoalState.WAITING_FOR_DATE, 1),
    ],
)
def test_handle_waiting_for_worker_any_mode_retry_without_goals(precond_state, expected_waiting_for_count):
    """
    In ANY mode, retry with precondition_goals=[] should wait for any not-achieved precondition.
    If all preconditions are achieved, we retry immediately.
    """
    precond = GoalFactory(state=precond_state)
    goal = GoalFactory(
        state=GoalState.WAITING_FOR_WORKER,
        handler='os.path.join',
        preconditions_mode=PreconditionsMode.ANY,
        waiting_for_count=0,
        waiting_for_not_achieved_count=1 if precond_state == GoalState.WAITING_FOR_DATE else 0,
    )
    goal.precondition_goals.add(precond)

    with mock.patch('os.path.join') as handler:
        handler.return_value = RetryMeLater(precondition_goals=[])
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    assert goal.waiting_for_count == expected_waiting_for_count


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'handler': 'os.path.join',
    'preconditions_mode': PreconditionsMode.ANY,
    'waiting_for_count': 0,
    'waiting_for_not_achieved_count': 1,
}], indirect=True)
def test_handle_waiting_for_worker_any_mode_all_done(goal):
    """
    In ANY preconditions mode, we can achieve the goal even if some preconditions are not met.
    """
    achieved_precond = GoalFactory(state=GoalState.ACHIEVED)
    waiting_precond = GoalFactory(state=GoalState.WAITING_FOR_DATE)
    goal.precondition_goals.add(achieved_precond, waiting_precond)

    with mock.patch('os.path.join') as handler:
        handler.return_value = AllDone()
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.ACHIEVED
    assert goal.waiting_for_count == 0
    assert goal.waiting_for_not_achieved_count == 1


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'state': GoalState.WAITING_FOR_WORKER,
    'handler': 'os.path.join',
    'preconditions_mode': PreconditionsMode.ANY,
    'waiting_for_count': 0,
    'waiting_for_not_achieved_count': 1,
}], indirect=True)
@pytest.mark.parametrize(
    ('local_precond_state', 'expected_waiting_for_count'),
    [
        (GoalState.ACHIEVED, 1),
        (GoalState.WAITING_FOR_DATE, 0),
    ],
)
def test_handle_waiting_for_worker_any_mode_retry_with_stale_goal_state(goal, local_precond_state, expected_waiting_for_count):
    """
    In ANY preconditions mode, we retry immediately if the precondition goal was achieved during handler execution.
    """
    old_precond = GoalFactory(state=GoalState.WAITING_FOR_DATE)
    goal.precondition_goals.add(old_precond)

    new_precond = GoalFactory(state=GoalState.ACHIEVED)
    new_precond.state = local_precond_state  # simulate we have stale state

    with mock.patch('os.path.join') as handler:
        handler.return_value = RetryMeLater(precondition_goals=[new_precond])
        handle_waiting_for_worker()

    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    assert goal.waiting_for_count == expected_waiting_for_count
    assert goal.waiting_for_not_achieved_count == 1


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('failure_mode', 'expected_state', 'progress_count'),
    [
        (PreconditionFailureBehavior.PROCEED, GoalState.ACHIEVED, 1),
        (PreconditionFailureBehavior.BLOCK, GoalState.NOT_GOING_TO_HAPPEN_SOON, 0),
    ],
)
@override_settings(GOALS_GIVE_UP_AT=1)
def test_proceed_at_failure_mode_do_work(failure_mode, expected_state, progress_count):
    precond = schedule(fail)
    goal = schedule(
        noop,
        precondition_goals=[precond],
        precondition_failure_behavior=failure_mode,
    )
    worker(once=True)
    precond.refresh_from_db()
    assert precond.state == GoalState.GIVEN_UP
    goal.refresh_from_db()
    assert goal.state == expected_state
    assert goal.progress.count() == progress_count


def noop(goal):  # pylint: disable=unused-argument
    return AllDone()


def fail(goal):  # pylint: disable=unused-argument
    raise Exception('I failed!')


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{'state': GoalState.WAITING_FOR_WORKER}], indirect=True)
def test_handle_waiting_for_worker_max_progress_exceeded(goal, settings):
    dependent_goal = GoalFactory(
        waiting_for_failed_count=122,
        precondition_goals=[goal],
    )

    settings.GOALS_MAX_PROGRESS_COUNT = 1
    with mock.patch('django_goals.models.follow_instructions') as follow_instructions:
        follow_instructions.return_value = RetryMeLater()
        handle_waiting_for_worker()
    goal.refresh_from_db()
    assert goal.state == GoalState.GIVEN_UP
    assert goal.progress.count() == 1

    # dependent goal is updated
    dependent_goal.refresh_from_db()
    assert dependent_goal.waiting_for_failed_count == 123


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize(
    ('goal', 'expected_state'),
    [
        ({
            'state': GoalState.WAITING_FOR_PRECONDITIONS,
            'waiting_for_count': 0,
        }, GoalState.WAITING_FOR_WORKER),
        ({
            'state': GoalState.WAITING_FOR_PRECONDITIONS,
            'waiting_for_count': 1,
        }, GoalState.WAITING_FOR_PRECONDITIONS),
        ({
            'state': GoalState.WAITING_FOR_PRECONDITIONS,
            'waiting_for_count': -1,
        }, GoalState.WAITING_FOR_WORKER),
        ({
            'state': GoalState.WAITING_FOR_DATE,
            'waiting_for_count': 0,
        }, GoalState.WAITING_FOR_DATE),
    ],
    indirect=['goal'],
)
def test_handle_waiting_for_preconditions(goal, expected_state, get_notifications):
    listen_goal_waiting_for_worker()

    handle_waiting_for_preconditions()

    goal.refresh_from_db()
    assert goal.state == expected_state

    # notification was sent accordingly
    notifications = get_notifications()
    if expected_state == GoalState.WAITING_FOR_WORKER:
        assert len(notifications) == 1
        notification = notifications[0]
        assert notification.channel == 'goal_waiting_for_worker'
        assert notification.payload == str(goal.id)
    else:
        assert not notifications


def trigger_database_error(goal):
    Goal.objects.create(id=goal.id)  # violates unique constraint


@pytest.mark.django_db(transaction=True)
def test_transaction_error_in_goal():
    goal = schedule(trigger_database_error)
    trasitions_count, progress_count = worker_turn(timezone.now())
    assert trasitions_count == 1
    assert progress_count == 1  # we called a handler and we must report it
    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    # progress record is created
    assert goal.progress.exists()
    assert thread_local.current_goal is None


def use_lots_of_memory(goal):  # pylint: disable=unused-argument
    _unused = b'x' * 1024 * 1024 * 128  # 128 MiB  # noqa
    return AllDone()


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('memory_limit', 'expected_success'),
    [
        (None, True),
        (1, False),
        (128, False),
        (256, True),
    ],
)
def test_memory_limit(settings, memory_limit, expected_success):
    settings.GOALS_MEMORY_LIMIT_MIB = memory_limit
    # simulate we have some memory allocated outside of the goal handler
    _unused = b'x' * 1024 * 1024 * 2  # 2 MiB  # noqa
    goal = schedule(use_lots_of_memory)
    worker_turn(timezone.now())
    goal.refresh_from_db()
    expected_state = GoalState.ACHIEVED if expected_success else GoalState.WAITING_FOR_DATE
    assert goal.state == expected_state
    progress = goal.progress.get()
    assert progress.success == expected_success


def take_too_long(goal):  # pylint: disable=unused-argument
    time.sleep(2)
    return AllDone()


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('time_limit', 'expected_success'),
    [
        (None, True),
        (1, False),
        (3, True),
    ],
)
def test_time_limit(settings, time_limit, expected_success):
    settings.GOALS_TIME_LIMIT_SECONDS = time_limit
    goal = schedule(take_too_long)
    worker_turn(timezone.now())
    goal.refresh_from_db()
    expected_state = GoalState.ACHIEVED if expected_success else GoalState.WAITING_FOR_DATE
    assert goal.state == expected_state
    progress = goal.progress.get()
    assert progress.success == expected_success


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('days_ago', 'state', 'expect_deleted'),
    [
        (31, GoalState.ACHIEVED, True),
        (1, GoalState.ACHIEVED, False),
        (31, GoalState.WAITING_FOR_WORKER, False),
        (31, GoalState.GIVEN_UP, False),
    ],
)
def test_old_achieved_goal_is_deleted(days_ago, state, expect_deleted):
    now = timezone.now()
    goal = GoalFactory(
        state=state,
        created_at=now - timezone.timedelta(days=days_ago),
    )
    GoalProgressFactory(goal=goal)
    dependent_goal = GoalFactory(precondition_goals=[goal])

    worker_turn(now)

    exists_after = Goal.objects.filter(id=goal.id).exists()
    assert exists_after is not expect_deleted

    # dependecy is removed
    dependent_goal.refresh_from_db()
    assert dependent_goal.precondition_goals.exists() is not expect_deleted


@pytest.mark.django_db
def test_protected_old_achieved_goal():
    now = timezone.now()
    goal = GoalFactory(
        state=GoalState.ACHIEVED,
        created_at=now - timezone.timedelta(days=31),
    )
    GoalRelatedModel.objects.create(goal=goal)

    # check we can't delete the goal
    with pytest.raises(models.ProtectedError):
        goal.delete()

    # worker turn doesn't crash, but emits a warning
    with mock.patch('django_goals.models.logger.warning') as warning:
        worker_turn(now)
    assert warning.call_count == 1
    assert 'old goals' in warning.call_args[0][0]
    assert 'protected' in str(warning.call_args[0][1])


def schedule_another(goal):
    schedule(schedule_another, blocked=True)
    return AllDone()


@pytest.mark.django_db
def test_deadline_is_inherited():
    now = timezone.now()
    goal = schedule(schedule_another, deadline=now + timezone.timedelta(days=1))
    worker_turn(now)
    another_goal = Goal.objects.exclude(id=goal.id).get()
    assert another_goal.deadline == goal.deadline



from django.core.management.base import BaseCommand

from django_goals.blocking_worker import worker


class Command(BaseCommand):
    help = 'Run the blocking worker'

    def handle(self, *args, **options):
        worker()



# Generated by Django 5.1.3 on 2024-12-11 18:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0005_goal_goals_waiting_for_precond_idx'),
    ]

    operations = [
        migrations.AddField(
            model_name='goal',
            name='waiting_for_failed_count',
            field=models.IntegerField(default=0, help_text='Number of precondition goals that failed.'),
        ),
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(condition=models.Q(('state', 'waiting_for_preconditions')), fields=['waiting_for_failed_count'], name='goals_waiting_for_failed_idx'),
        ),
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(condition=models.Q(('state', 'not_going_to_happen_soon')), fields=['waiting_for_failed_count'], name='goals_unblocking_idx'),
        ),
    ]



import logging

from django.db import connection

from .models import handle_waiting_for_worker
from .notifications import listen_goal_waiting_for_worker


logger = logging.getLogger(__name__)


def worker():
    '''
    This worker is a blocking worker that listens for notifications on the
    goal_waiting_for_worker channel. It will then handle the waiting for worker
    jobs. This worker will run indefinitely until it is stopped.
    '''
    logger.info("Blocking worker started, registering listener (goal_waiting_for_worker)")
    listen_goal_waiting_for_worker()

    logger.info("Executing work ready before we were listening")
    while True:
        did_a_thing = handle_waiting_for_worker()
        if not did_a_thing:
            break

    logger.info("Handling notifications")
    pg_conn = connection.connection
    for _ in pg_conn.notifies():
        # We might pick a different job than the one that was notified.
        # This is okay, because there are as many (or more) notifications as there are jobs.
        handle_waiting_for_worker()

    logger.info("Blocking worker exiting now")



# Generated by Django 5.2 on 2025-04-15 08:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_goals", "0009_goal_waiting_for_not_achieved_count"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="goal",
            name="goals_waiting_for_failed_idx",
        ),
        migrations.AddField(
            model_name="goal",
            name="precondition_failure_behavior",
            field=models.CharField(
                choices=[
                    ("block", "Do not proceed if preconditions fail"),
                    (
                        "proceed",
                        "Proceed with goal execution even if preconditions fail",
                    ),
                ],
                default="block",
                max_length=10,
            ),
        ),
        migrations.AddIndex(
            model_name="goal",
            index=models.Index(
                condition=models.Q(
                    ("precondition_failure_behavior", "block"),
                    ("state", "waiting_for_preconditions"),
                ),
                fields=["waiting_for_failed_count"],
                name="goals_waiting_for_failed_idx",
            ),
        ),
        migrations.AddConstraint(
            model_name="goal",
            constraint=models.CheckConstraint(
                condition=models.Q(
                    ("precondition_failure_behavior__in", ["block", "proceed"])
                ),
                name="goals_precondition_failure_behavior",
            ),
        ),
    ]



import time

from django.core.management.base import BaseCommand

from django_goals.models import AllDone, schedule, wait


def pursue(goal):
    return AllDone()


class Command(BaseCommand):
    help = 'Set an example goal'

    def handle(self, *args, **options):
        start_time = time.monotonic()
        goal = schedule(pursue, listen=True)
        print('Goal scheduled', goal.id)
        notification = wait()
        end_time = time.monotonic()
        print('Goal done', notification, 'in', end_time - start_time, 'seconds')



# Generated by Django 5.1.3 on 2024-12-08 00:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0003_remove_goal_goals_waiting_for_worker_idx_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='goal',
            name='waiting_for_count',
            field=models.IntegerField(
                default=0,
                help_text='Number of precondition goals that must finish before this goal can be pursued.',
            ),
        ),
    ]



import pytest

from .blocking_worker import listen_goal_waiting_for_worker
from .models import schedule


def noop():
    pass


@pytest.mark.django_db(transaction=True)
def test_schedule_notifies(get_notifications):
    listen_goal_waiting_for_worker()

    goal = schedule(noop)

    notifications = get_notifications()
    assert len(notifications) == 1
    notification = notifications[0]
    assert notification.channel == 'goal_waiting_for_worker'
    assert notification.payload == str(goal.id)



from datetime import timedelta

import pytest
from django.core.management import call_command
from django.utils.timezone import now

from django_goals.models import AllDone, GoalState, schedule


def achieve(goal):
    return AllDone()


@pytest.mark.django_db(transaction=True)
def test_achieving_goals():
    goal = schedule(achieve)
    call_command('goals_threaded_worker', threads=['2'], once=True)
    goal.refresh_from_db()
    assert goal.state == GoalState.ACHIEVED


@pytest.mark.parametrize(
    ('threads_spec', 'expected_goal_state'),
    [
        (['1:1m'], GoalState.WAITING_FOR_WORKER),
        (['1:1m', '1:2d'], GoalState.ACHIEVED),
        (['1:2d', '1:1m'], GoalState.ACHIEVED),
        (['1:0s'], GoalState.WAITING_FOR_WORKER),
    ],
)
@pytest.mark.django_db(transaction=True)
def test_deadline_horizon(threads_spec, expected_goal_state):
    """ Worker should pick up goals with deadline within the deadline horizon """
    goal = schedule(achieve, deadline=now() + timedelta(days=1))
    call_command('goals_threaded_worker', threads=threads_spec, once=True)
    goal.refresh_from_db()
    assert goal.state == expected_goal_state



from unittest import mock

import pytest
from django.db import connection

from .factories import GoalFactory


@pytest.fixture(name='goal')
def goal_fixture(request):
    return GoalFactory(
        **getattr(request, 'param', {}),
    )


@pytest.fixture(name='get_notifications')
def get_notifications_fixture():
    handler = mock.Mock()
    connection.ensure_connection()
    pg_conn = connection.connection
    pg_conn.add_notify_handler(handler)

    def _get_notifications():
        return [
            call[0][0]
            for call in handler.call_args_list
        ]
    return _get_notifications



import pytest
from django.urls import reverse

from django_goals.models import GoalState


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('goal', 'action', 'expected_state'),
    [
        (
            {'state': GoalState.WAITING_FOR_WORKER},
            'block',
            GoalState.BLOCKED,
        ),
        (
            {'state': GoalState.BLOCKED},
            'unblock_retry',
            GoalState.WAITING_FOR_DATE,
        ),
        (
            {'state': GoalState.GIVEN_UP},
            'unblock_retry',
            GoalState.WAITING_FOR_DATE,
        ),
        (
            {'state': GoalState.NOT_GOING_TO_HAPPEN_SOON},
            'unblock_retry',
            GoalState.WAITING_FOR_DATE,
        ),
    ],
    indirect=['goal'],
)
def test_state_actions(admin_client, goal, action, expected_state):
    response = admin_client.post(reverse(
        'admin:django_goals_goal_actions',
        args=[goal.pk, action],
    ))
    assert response.status_code == 302
    goal.refresh_from_db()
    assert goal.state == expected_state



# Generated by Django 5.0.6 on 2024-06-10 16:05

import uuid

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Goal',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('state', models.CharField(choices=[
                    ('blocked', 'Blocked'),
                    ('waiting_for_date', 'Waiting For Date'),
                    ('waiting_for_preconditions', 'Waiting For Preconditions'),
                    ('waiting_for_worker', 'Waiting For Worker'),
                    ('achieved', 'Achieved'),
                    ('given_up', 'Given Up'),
                    ('not_going_to_happen_soon', 'Not Going To Happen Soon'),
                ], db_index=True, default='waiting_for_date', max_length=30)),
                ('handler', models.CharField(max_length=100)),
                ('instructions', models.JSONField(null=True)),
                ('precondition_date', models.DateTimeField(default=django.utils.timezone.now, help_text='Goal will not be pursued before this date.')),
                ('created_at', models.DateTimeField(db_index=True, default=django.utils.timezone.now)),
            ],
            options={
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='GoalDependency',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('dependent_goal', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='dependencies', to='django_goals.goal')),
                ('precondition_goal', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='dependents', to='django_goals.goal')),
            ],
            options={
                'unique_together': {('dependent_goal', 'precondition_goal')},
            },
        ),
        migrations.AddField(
            model_name='goal',
            name='precondition_goals',
            field=models.ManyToManyField(blank=True, related_name='dependent_goals', through='django_goals.GoalDependency', to='django_goals.goal'),
        ),
        migrations.CreateModel(
            name='GoalProgress',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('success', models.BooleanField()),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('time_taken', models.DurationField(null=True)),
                ('goal', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='progress', to='django_goals.goal')),
            ],
            options={
                'ordering': ('goal', '-created_at'),
            },
        ),
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(condition=models.Q(('state', 'waiting_for_date')), fields=['precondition_date'], name='goals_waiting_for_date_idx'),
        ),
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(condition=models.Q(('state', 'waiting_for_worker')), fields=['precondition_date'], name='goals_waiting_for_worker_idx'),
        ),
    ]



# Generated by Django 5.0.7 on 2024-09-10 16:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='goalprogress',
            name='message',
            field=models.CharField(blank=True, max_length=200),
        ),
    ]



# Generated by Django 5.1.4 on 2025-01-11 15:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0008_remove_goal_goals_waiting_for_precond_idx_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='goal',
            name='waiting_for_not_achieved_count',
            field=models.IntegerField(default=0, help_text='Number of precondition goals that are not achieved yet.'),
        ),
    ]



import datetime

import pytest

from .factories import GoalFactory
from .models import (
    GoalState, PreconditionFailureBehavior, PreconditionsMode,
    handle_unblocked_goals, schedule, unblock_retry_goal,
)


@pytest.mark.django_db
@pytest.mark.parametrize(
    'goal',
    [{'state': GoalState.GIVEN_UP}],
    indirect=True,
)
def test_retry(goal):
    unblock_retry_goal(goal.id)
    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{'state': GoalState.GIVEN_UP}], indirect=True)
def test_retry_dependent_on(goal):
    next_goal = GoalFactory(
        state=GoalState.NOT_GOING_TO_HAPPEN_SOON,
        precondition_goals=[goal],
    )
    unblock_retry_goal(goal.id)
    handle_unblocked_goals()
    next_goal.refresh_from_db()
    assert next_goal.state == GoalState.WAITING_FOR_DATE


def noop(goal):  # pylint: disable=unused-argument
    pass


@pytest.mark.django_db
def test_schedule_updates_deadline():
    now = datetime.datetime(2024, 11, 6, 11, 41, 0, tzinfo=datetime.timezone.utc)
    goal_a = GoalFactory(deadline=now)
    goal_b = GoalFactory(precondition_goals=[goal_a])
    schedule(
        noop,
        deadline=now - datetime.timedelta(minutes=1),
        precondition_goals=[goal_b],
    )
    goal_a.refresh_from_db()
    assert goal_a.deadline == now - datetime.timedelta(minutes=1)


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('goal', 'expected_waiting_for', 'expected_waiting_for_failed_count'),
    [
        ({'state': GoalState.WAITING_FOR_WORKER}, 1, 0),
        ({'state': GoalState.ACHIEVED}, 0, 0),
        ({'state': GoalState.GIVEN_UP}, 1, 1),
        ({'state': GoalState.NOT_GOING_TO_HAPPEN_SOON}, 1, 1),
    ],
    indirect=['goal'],
)
@pytest.mark.parametrize('mode', [PreconditionsMode.ALL, PreconditionsMode.ANY])
def test_schedule_updates_waiting_for_count(goal, expected_waiting_for, expected_waiting_for_failed_count, mode):
    next_goal = schedule(noop, precondition_goals=[goal], preconditions_mode=mode)
    assert next_goal.waiting_for_count == expected_waiting_for
    assert next_goal.waiting_for_failed_count == expected_waiting_for_failed_count
    assert next_goal.preconditions_mode == mode


@pytest.mark.django_db
@pytest.mark.parametrize(
    'failure_mode',
    [
        PreconditionFailureBehavior.PROCEED,
        PreconditionFailureBehavior.BLOCK,
    ],
)
def test_schedule_any_mode_caps_waiting_for(failure_mode):
    preconds = GoalFactory.create_batch(2, state=GoalState.WAITING_FOR_WORKER)
    next_goal = schedule(
        noop,
        precondition_goals=preconds,
        preconditions_mode=PreconditionsMode.ANY,
        precondition_failure_behavior=failure_mode,
    )
    assert next_goal.waiting_for_count == 1
    assert next_goal.precondition_goals.count() == 2


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('failure_mode', 'expected_waiting_for_count'),
    [
        (PreconditionFailureBehavior.PROCEED, 0),
        (PreconditionFailureBehavior.BLOCK, 1),
    ],
)
def test_schedule_failed_precond(failure_mode, expected_waiting_for_count):
    failed_goal = GoalFactory(
        state=GoalState.GIVEN_UP,
    )
    goal = schedule(
        noop,
        precondition_goals=[failed_goal],
        precondition_failure_behavior=failure_mode,
    )
    assert goal.state == GoalState.WAITING_FOR_PRECONDITIONS
    assert goal.waiting_for_count == expected_waiting_for_count
    assert goal.waiting_for_failed_count == 1
    assert goal.waiting_for_not_achieved_count == 1


@pytest.mark.django_db
@pytest.mark.parametrize('blocked', [True, False])
def test_schedule_blocked(blocked):
    goal = GoalFactory(state=GoalState.WAITING_FOR_WORKER)
    next_goal = schedule(noop, precondition_goals=[goal], blocked=blocked)
    assert next_goal.state == (
        GoalState.BLOCKED if blocked
        else GoalState.WAITING_FOR_PRECONDITIONS
    )



# Generated by Django 5.1.4 on 2024-12-11 23:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0006_goal_waiting_for_failed_count_and_more'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(condition=models.Q(('state', 'achieved')), fields=['created_at'], name='goals_achieved_idx'),
        ),
    ]



import uuid

from django.core.management.base import BaseCommand
from django.db import transaction

from django_goals.models import Goal, GoalState, unblock_retry_goal


class Command(BaseCommand):
    """
    Retry goals that were previously given up due to multiple failures.
    """
    help = 'Retry all goals that were previously marked as GIVEN_UP'

    def add_arguments(self, parser):
        parser.add_argument(
            '--limit',
            type=int,
            help='Maximum number of goals to retry (default: no limit)'
        )

    def handle(self, *args, **options):
        limit = options.get('limit')
        count = retry_all_given_up(self, limit)
        self.stdout.write(self.style.SUCCESS(f'Retried {count} goals'))


def retry_all_given_up(command, limit=None):
    """Retry all given up goals, optionally limited to a maximum count."""
    goal_id = uuid.UUID(int=0)
    count = 0

    while True:
        if limit is not None and count >= limit:
            command.stdout.write(f'Reached limit of {limit} goals')
            break

        goal_id = retry_next_given_up_goal(goal_id, command)

        if not goal_id:
            break

        count += 1

        # Move to the next goal ID
        goal_id = uuid.UUID(int=goal_id.int + 1)

    return count


@transaction.atomic
def retry_next_given_up_goal(goal_id, command):
    """
    Find and retry the next given up goal with ID >= goal_id.
    Returns the ID of the processed goal, or None if no eligible goal was found.
    """
    goal = Goal.objects.filter(
        id__gte=goal_id,
        state=GoalState.GIVEN_UP,
    ).order_by('id').select_for_update(
        no_key=True,
        skip_locked=True,
    ).first()

    if not goal:
        return None

    unblock_retry_goal(goal.id)
    command.stdout.write(f'Retried goal {goal.id}')

    return goal.id



import factory

from .models import Goal, GoalProgress


class GoalFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Goal
        skip_postgeneration_save = True

    @factory.post_generation
    def precondition_goals(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            self.precondition_goals.set(extracted)


class GoalProgressFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = GoalProgress

    goal = factory.SubFactory(GoalFactory)
    success = True
